"""Cosmology applications"""

from mcfit.mcfit import mcfit
from mcfit import kernels
from numpy import pi


__all__ = ['P2xi', 'xi2P', 'TophatVar', 'GaussVar']


class P2xi(mcfit):
    """Power spectrum to correlation function.

    Parameters
    ----------
    k : see `x` in :class:`mcfit.mcfit`
    l : int
        order
    n : int
        to generalize correlation function with extra power law factor
        :math:`k^n` in the integrand. If not None, the phase factor is ignored.
        The tilt parameter `q` is automatically adjusted (to `q+n`) based on
        the provided value

    See :class:`mcfit.mcfit`
    """
    def __init__(self, k, l=0, n=None, deriv=0, q=1.5, **kwargs):
        self.l = l
        MK = kernels.Mellin_SphericalBesselJ(l, deriv)

        if n is None:
            phase = (-1 if l & 2 else 1) * (1j if l & 1 else 1)  # i^l
            n = 0
        else:
            phase = 1

        mcfit.__init__(self, k, MK, q+n, **kwargs)
        self.prefac *= self.x**(3+n) / (2*pi)**1.5
        self.postfac *= phase


class xi2P(mcfit):
    """Correlation function to power spectrum, also radial profile to its
    Fourier transform.

    Parameters
    ----------
    r : see `x` in :class:`mcfit.mcfit`
    l : int
        order

    See :class:`mcfit.mcfit`
    """
    def __init__(self, r, l=0, deriv=0, q=1.5, **kwargs):
        self.l = l
        MK = kernels.Mellin_SphericalBesselJ(l, deriv)
        mcfit.__init__(self, r, MK, q, **kwargs)
        self.prefac *= self.x**3
        phase = (-1 if l & 2 else 1) * (1j if l & 1 else 1)  # i^l
        self.postfac *= (2*pi)**1.5 / phase


class C2w(mcfit):
    """Angular power spectrum to correlation function.

    Parameters
    ----------
    ell : see `x` in :class:`mcfit.mcfit`
    nu : int
        order

    See :class:`mcfit.mcfit`
    """
    def __init__(self, ell, nu=0, deriv=0, q=1, **kwargs):
        self.nu = nu
        MK = kernels.Mellin_BesselJ(nu, deriv)
        mcfit.__init__(self, ell, MK, q, **kwargs)
        self.prefac *= self.x**2 / (2*pi)


class w2C(mcfit):
    """Angular correlation function to power spectrum.

    Parameters
    ----------
    theta : see `x` in :class:`mcfit.mcfit`
    nu : int
        order

    See :class:`mcfit.mcfit`
    """
    def __init__(self, theta, nu=0, deriv=0, q=1, **kwargs):
        self.nu = nu
        MK = kernels.Mellin_BesselJ(nu, deriv)
        mcfit.__init__(self, theta, MK, q, **kwargs)
        self.prefac *= self.x**2
        self.postfac *= 2*pi


class TophatVar(mcfit):
    r"""Variance in a top-hat window.

    Parameters
    ----------
    k : see `x` in :class:`mcfit.mcfit`

    Examples
    --------
    To compute :math:`\sigma_8` of a linear power spectrum :math:`P(k)`
    >>> R, var = TophatVar(k, lowring=True)(P, extrap=True)
    >>> from scipy.interpolate import CubicSpline
    >>> varR = CubicSpline(R, var)
    >>> sigma8 = numpy.sqrt(varR(8))

    See :class:`mcfit.mcfit`
    """
    def __init__(self, k, deriv=0, q=1.5, **kwargs):
        MK = kernels.Mellin_TophatSq(3, deriv)
        mcfit.__init__(self, k, MK, q, **kwargs)
        self.prefac *= self.x**3 / (2 * pi**2)


class GaussVar(mcfit):
    """Variance in a Gaussian window.

    Parameters
    ----------
    k : see `x` in :class:`mcfit.mcfit`

    See :class:`mcfit.mcfit`
    """
    def __init__(self, k, deriv=0, q=1.5, **kwargs):
        MK = kernels.Mellin_GaussSq(deriv)
        mcfit.__init__(self, k, MK, q, **kwargs)
        self.prefac *= self.x**3 / (2 * pi**2)



import numpy as np
from numpy.testing import assert_allclose
from mcfit.transforms import *


def test_matrix():
    N = 81
    x = np.logspace(-3, 3, num=N, endpoint=False)
    F = 1 / (1 + x*x)**1.5

    H1 = Hankel(x, nu=0, q=1, N=N)
    y, G = H1(F)

    a1, b1, C1 = H1.matrix(full=False)
    M1 = H1.matrix(full=True)

    assert_allclose(a1.ravel() * (C1 @ (b1 * F)), G, rtol=1e-10, atol=0)
    assert_allclose(M1 @ F, G, rtol=1e-10, atol=0)

    H2 = Hankel(y, nu=0, q=1, N=N)

    a2, b2, C2 = H2.matrix(full=False)
    M2 = H2.matrix(full=True)

    assert_allclose(C1 @ C2, np.eye(N), rtol=0, atol=1e-14)
    assert_allclose(C2 @ C1, np.eye(N), rtol=0, atol=1e-14)
    assert_allclose(M1 @ M2, np.eye(N), rtol=0, atol=1e-9)
    assert_allclose(M2 @ M1, np.eye(N), rtol=0, atol=1e-9)


def test_pad():
    # _x_, _y_, extrap pad, and unpad
    x = np.logspace(-3, 3, num=6, endpoint=False)
    _x_ = np.logspace(-6, 6, num=13, endpoint=True)
    y = np.logspace(-3, 3, num=6, endpoint=False)
    _y_ = np.logspace(-7, 5, num=13, endpoint=True)

    H = Hankel(x, N=13, xy=1)

    assert_allclose(H._x_, _x_)
    assert_allclose(H._y_, _y_)

    assert_allclose(H._pad(x, 0, True, False), _x_)
    assert_allclose(H._pad(y, 0, True, True), _y_)

    assert_allclose(H._unpad(_x_, 0, False), x)
    assert_allclose(H._unpad(_y_, 0, True), y)

    # zero pad and axis
    a = b = np.ones((6, 6))
    _a_ = np.zeros((13, 6))
    _a_[3:9, :] = 1
    _b_ = np.zeros((6, 13))
    _b_[:, 4:10] = 1

    assert_allclose(H._pad(a, 0, False, False), _a_)
    assert_allclose(H._pad(b, 1, False, True), _b_)
